(* s24path.ml     Drawing paths
 *
 * Copyright (c) 2022 Psellos   http://psellos.com
 *
 * Licensed under the MIT license:
 *     http://opensource.org/licenses/MIT
 *)

type bzp_elem =
| BZBegin
| BZMove of S24look.point
| BZLine of S24look.point
| BZCurve of S24look.point * S24look.point * S24look.point (* Cp1 Cp2 Endp *)
| BZClose

type bzpath = bzp_elem list


let bzp_pmap pfun bzpath =
    let efun = function
    | BZMove p -> BZMove (pfun p)
    | BZLine p -> BZLine (pfun p)
    | BZCurve (c1, c2, p) -> BZCurve (pfun c1, pfun c2, pfun p)
    | (BZBegin | BZClose) as e -> e
    in
    List.map efun bzpath


let bzp_iter efun bzpath =
    List.iter efun bzpath


let bbox bzpath =
    (* Return bounding box for the path. We assume the path starts with
     * BZMove (so initial position is immaterial) and that its curves
     * aren't too erratic (they stay inside the rectangle bounded by
     * start and end points). These are true for the paths we work with
     * here.
     *)
    let bbmax (xmin, xmax, ymin, ymax  as bb) elem =
        match elem with
        | BZMove (x, y)
        | BZLine (x, y)
        | BZCurve (_, _, (x, y)) ->
            (min x xmin, max x xmax, min y ymin, max y ymax)
        | _ -> bb
    in
    let (xmin, xmax, ymin, ymax) =
        List.fold_left bbmax (1_000_000.0, 0.0, 1_000_000.0, 0.0) bzpath
    in
    (xmin, ymin, xmax -. xmin, ymax -. ymin)


let bzp_scale bzpath (x, y, w, h) =
    let (bx, by, bw, bh) = bbox bzpath in
    let s, tx, ty =
        if (w /. bw) < (h /. bh) then
            let s = w /. bw in
            let tx = x -. s *. bx in
            let ty = y -. s *. by -. (s *. bh -. h) /. 2.0 in
            s, tx, ty
        else
            let s = h /. bh in
            let tx = x -. s *. bx -. (s *. bw -. w) /. 2.0 in
            let ty = y +. h -. s *. (by +. bh) in
            s, tx, ty
    in
    let fx x = s *. x +. tx in
    let fy y = s *. y +. ty in
    let fp (x, y) = (fx x, fy y) in
    bzp_pmap fp bzpath


let draw ctx2d bzpath =
    let draw1 = function
    | BZBegin -> ctx2d##beginPath
    | BZMove (x, y) -> ctx2d##moveTo x y
    | BZLine (x, y) -> ctx2d##lineTo x y
    | BZCurve ((c1x, c1y), (c2x, c2y), (x, y)) ->
        ctx2d##bezierCurveTo c1x c1y c2x c2y x y
    | BZClose -> ctx2d##closePath
    in
    List.iter draw1 bzpath;
    ctx2d##fill
